/*
    pcb-rnd-3dview - watch a file for changes and run a viewer bg-process
    Copyright (C) 2021  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://www.repo.hu/projects/pcb-rnd-3dview
    VCS: svn://svn.repo.hu/pcb-rnd-3dview/trunk
    Contact: http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <sys/types.h>
#include <unistd.h>

#include "projview.h"

#define PIPE_RD 0
#define PIPE_WR 1

void projview_reset(projview_t *ctx)
{
	ctx->ax = ctx->ay = ctx->az = 0;
	ctx->ocx = ctx->ocy = ctx->ocz = 0;
}

void projview_init(projview_t *ctx)
{
	int pipe_v2p[2], pipe_p2v[2];

	projview_reset(ctx);
	ctx->zoom = 100;

	if ((pipe(pipe_v2p) != 0) || (pipe(pipe_p2v) != 0)) {
		fprintf(stderr, "Failed to create pipes for projector\n");
		exit(1);
	}

	if (fork() == 0) {
		close(pipe_v2p[PIPE_WR]);
		close(pipe_p2v[PIPE_RD]);
		close(0);
		dup(pipe_v2p[PIPE_RD]);
		close(pipe_v2p[PIPE_RD]);
		close(1);
		dup(pipe_p2v[PIPE_WR]);
		close(pipe_p2v[PIPE_WR]);

		/* child process */
		execlp("projector", "projector", "-e", NULL);
		fprintf(stderr, "Failed to execute projector\n");
		exit(1);
	}
	close(pipe_v2p[PIPE_RD]);
	close(pipe_p2v[PIPE_WR]);

	ctx->prj_wfd = pipe_v2p[PIPE_WR];
	ctx->prj_rfd = pipe_p2v[PIPE_RD];
	ctx->prj_rf = fdopen(ctx->prj_rfd, "r");
	if (ctx->prj_rf == NULL) {
		fprintf(stderr, "Failed to fdopen() projector read\n");
		exit(1);
	}
	ctx->prj_wf = fdopen(ctx->prj_wfd, "w");
	if (ctx->prj_wf == NULL) {
		fprintf(stderr, "Failed to fdopen() projector write\n");
		exit(1);
	}

}

static int projview_load_(projview_t *ctx, FILE *f)
{
	char line[1024];
	while(fgets(line, sizeof(line), f) != NULL)
		fputs(line, ctx->prj_wf);
	fflush(ctx->prj_wf);
	return 0;
}

int projview_load_file(projview_t *ctx, const char *fn)
{
	int res;
	FILE *f;

	if (ctx->prj_wfd < 0)
		return -1;

	f = fopen(fn, "r");
	if (f == NULL)
		return -1;

	res = projview_load_(ctx, f);

	fclose(f);
	return res;
}

int projview_load_proc(projview_t *ctx, const char *cmd)
{
	int res;
	FILE *f;

	if (ctx->prj_wfd < 0)
		return -1;

	f = popen(cmd, "r");
	if (f == NULL)
		return -1;

	res = projview_load_(ctx, f);

	pclose(f);
	return res;
}

int projview_cam(projview_t *ctx)
{
	if (ctx->prj_wfd < 0)
		return -1;

	fprintf(ctx->prj_wf, "obj \"%s\" mat  rotx %f * roty %f * rotz %f * shift %f %f %f\n",
		ctx->objname, ctx->ax, ctx->ay, ctx->az, ctx->ocx, ctx->ocy, 0.0);
	fprintf(ctx->prj_wf, "cam 0 0 %f 0 0 0\n", ctx->zoom);
	fprintf(ctx->prj_wf, "light 1000000 1000000 1000000");

	fprintf(ctx->prj_wf, "render\n");
	fflush(ctx->prj_wf);

	return 0;
}

int projview_tick(projview_t *ctx)
{
	double dx = ctx->pressx - ctx->motx;
	double dy = ctx->pressy - ctx->moty;

	if (ctx->press == 1) {
		ctx->az = ctx->pressaz + (double)dx + (double)dy;
		projview_cam(ctx);
	}
	else if (ctx->press == 2) {
		ctx->ocx = ctx->pressocx + (double)dy / ctx->zoom * 3.0;
		ctx->ocy = ctx->pressocy + (double)dx / ctx->zoom * 3.0;
		projview_cam(ctx);
	}
	else if (ctx->press == 3) {
		ctx->ax = ctx->pressax + (double)dx;
		ctx->ay = ctx->pressay + (double)dy;
		projview_cam(ctx);
	}
	return 0;
}

static void prj_press(projview_t *ctx, const char *args)
{
	sscanf(args, "%d %lf %lf", &ctx->press, &ctx->pressx, &ctx->pressy);

	ctx->pressax = ctx->ax;
	ctx->pressay = ctx->ay;
	ctx->pressaz = ctx->az;
	ctx->pressocx = ctx->ocx;
	ctx->pressocy = ctx->ocy;
	ctx->pressocz = ctx->ocz;

	if (ctx->press == 4) {
		ctx->zoom /= 1.3;
		projview_cam(ctx);
	}
	else if (ctx->press == 5) {
		ctx->zoom *= 1.3;
		projview_cam(ctx);
	}
}

static void prj_release(projview_t *ctx, const char *args)
{
	ctx->press = 0;
}

static void prj_motion(projview_t *ctx, const char *args)
{
	/* skip $2 */
	while(!isspace(*args)) args++;
	while(isspace(*args)) args++;

	sscanf(args, "%lf %lf", &ctx->motx, &ctx->moty);
}

static void prj_keyup(projview_t *ctx, const char *args)
{
	/* skip $2 */
	while(!isspace(*args)) args++;
	while(isspace(*args)) args++;

	switch(args[0]) {
		case 'q': exit(1);
		case '1': projview_reset(ctx); projview_cam(ctx); break;
		case '2': projview_reset(ctx); ctx->ay = 180; projview_cam(ctx); break;
		case '3': projview_reset(ctx); ctx->ax = 90; projview_cam(ctx); break;
		case '4': projview_reset(ctx); ctx->ay = 90; projview_cam(ctx); break;
		case '5': projview_reset(ctx); ctx->ax = -90; projview_cam(ctx); break;
		case '6': projview_reset(ctx); ctx->ay = -90; projview_cam(ctx); break;
	}
}

void projview_read(projview_t *ctx)
{
	char line[1024], *cmd, *args;

	if (fgets(line, sizeof(line), ctx->prj_rf) <= 0) {
		fclose(ctx->prj_rf);
		fclose(ctx->prj_wf);
		ctx->prj_rf = ctx->prj_wf = NULL;
		ctx->prj_rfd = ctx->prj_wfd = 0;
		return;
	}

	cmd = line;
	while(isspace(*cmd)) cmd++;
	args = cmd;
	while(!isspace(*args)) args++;
	*args = '\0';
	args++;
	while(isspace(*args)) args++;

	if (strcmp(cmd, "PRESS") == 0) prj_press(ctx, args);
	else if (strcmp(cmd, "RELEASE") == 0) prj_release(ctx, args);
	else if (strcmp(cmd, "MOTION") == 0) prj_motion(ctx, args);
	else if (strcmp(cmd, "KEYUP") == 0) prj_keyup(ctx, args);
}
